# meds/bots/irc.py
#
#

""" Internet Relay Chat Bot. """

from meds.utils.misc import split_txt, locked
from meds.utils.trace import get_exception
from meds.errors import EDISCONNECT
from meds.storage import Storage
from meds.object import Object
from meds.kernel import kernel
from meds.utils.join import j
from meds.event import Event
from meds.bots import Bot
from meds.cfg import cfg

from meds import __version__

import logging
import _thread
import select
import random
import socket
import queue
import time
import ssl
import re
import os

class IRC(Bot):

    """ A bot that connects to IRC networks. """

    cc = "!"
    default = ""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._resume = Object()
        self._lock = _thread.allocate_lock()
        self._outqueue = queue.Queue()
        self._sock = None
        self._buffer = []
        self._lastline = ""
        self._error = ""
        self._status = "start"
        self._nrconnect = 0
        self._last = time.time()
        self.poll = select.epoll()
        self.handlers = Object()
        self.started = False
        self.register("004", self.connected)
        self.register("ERROR", self.errored)
        self.register("366", self.h366)
        self.register("433", self.h433)
        self.register("513", self.h513)
        self.register("PING", self.pinged)
        self.register("PONG", self.ponged)
        self.register("QUIT", self.quited)
        self.register("INVITE", self.invited)
        self.register("PRIVMSG", self.privmsged)
        self.register("NOTICE", self.noticed)
        self.register("JOIN", self.joined)
        self.channels = []
        self.encoding = "utf-8"
        if "realname" not in self: self.realname = "meds"
        if "username" not in self: self.username = "meds"
        if "server" not in self: self.server = "localhost"
        if "port" not in self: self.port = 6667
        if "nick" not in self: self.nick = "meds"
        if "channel" in self: self.channels.append(self.channel)

    def register(self, key, value):
        """ register a handler callback. """
        self.handlers[key] = value

    def dispatch(self, event):
        """ dispatch an event onto a registered handler. """
        if event.cmnd in self.handlers: self.handlers[event.cmnd](event)

    def start(self):
        """ start IRC bot. """
        store = Storage()
        config = store.last("cfg", "irc")
        if not config: config = cfg.irc ; config.save()
        self.update(config)
        self.launch(self.push)
        if "channel" in self: self.channels.append(self.channel)
        if not cfg.main.resume: self.connecting()
        else: self.resume()
        self.launch(super().start)

    def stop(self, close=True):
        """ stop IRC bot. """
        logging.warn("# stop %s" % self.server)
        if close:
            self._status = ""
            #self.out("stop")
            self._outqueue.put(None)
            super().stop()
        self.kill("IRC")
  
    def resume(self):
        logging.warn("# resume %s" % self.server)
        kernel = Object().load(j(cfg.main.workdir, "kernel"))
        bots = kernel.fleet.get("IRC", None)
        for bot in bots:
            fd = int(bot["_resume"]["fd"])
            if cfg.irc.ipv6: self._oldsock = socket.fromfd(fd , socket.AF_INET6, socket.SOCK_STREAM)
            else: self._oldsock = socket.fromfd(fd, socket.AF_INET, socket.SOCK_STREAM)
            self.socket_cfg()
        self.announce("done")

    def events(self):
        for (fd, events) in self.poll.poll(-1): yield self.event()

    def socket_cfg(self):
        self._oldsock.settimeout(60.0)
        if not cfg.main.resume:
            logging.warn("# connect %s:%s" % (self.server, self.port or 6667))
            self._oldsock.connect((self.server, int(str(self.port or 6667))))
        self.blocking = True
        self._oldsock.setblocking(self.blocking)
        self._oldsock.settimeout(700.0)
        self.fsock = self._oldsock.makefile("r")
        if 'ssl' in self and self['ssl']: self._sock = ssl.wrap_socket(self._oldsock)
        else: self._sock = self._oldsock
        self._resume.fd = self._sock.fileno()
        os.set_inheritable(self._resume.fd, os.O_RDWR)
        self.add(self._sock)

    def push(self):
        """ loop to output txt to the server. """
        while self._status:
            args = self._outqueue.get()
            if not args or not self._status: break
            try: self.output(*args)
            except BrokenPipeError: time.sleep(2.0) ; self.connecting()
            except: logging.error(get_exception()) ; time.sleep(5)

    def event(self):
        """ read an IRC event. """
        if not self._buffer:
            try: self.some()
            except socket.timeout: self.some()
        line = self._buffer.pop(0)
        return self.parse(line.rstrip())

    def some(self):
        """ read the socket until data arrives. """
        if "ssl" in self and self.ssl: inbytes = self._sock.read()
        else: inbytes = self._sock.recv(512)
        txt = str(inbytes, self.encoding)
        if txt == "": raise EDISCONNECT()
        self._lastline += txt
        splitted = self._lastline.split("\r\n")
        for s in splitted[:-1]:
            self._buffer.append(s)
            if "PING" not in s or "PONG" not in s: logging.info("< %s" % s.strip())
        self._lastline = splitted[-1]

    def connecting(self):
        """ start a connecting loop until bot is connected. """
        self._status = "connect"
        self._nrconnect = 0
        while self._status:
            try: self.connect() ; break
            except Exception as ex:
                self._status = str(ex)
                self._nrconnect += 1
                time.sleep(self._nrconnect * 10)
        try: self.logon()
        except Exception as ex: logging.error("^ error %s" % str(ex))
        self._status = "ok"

    def connect(self):
        """ connect to the server. """
        self.stopped = False
        # create socket
        if "ipv6" in self and self.ipv6: self._oldsock = socket.socket(socket.AF_INET6, socket.SOCK_STREAM)
        else: self._oldsock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.server = self.bind() 
        self.socket_cfg()

    def logon(self, *args):
        """ logon the IRC network. """
        self.out("NICK %s" % self.nick or "meds")
        self.out("USER %s localhost %s :%s" % (self.username or "meds", self.server or "localhost", self.realname or "meds"))

    def bind(self):
        """ bind the socket to a hostname. """
        server = self.server
        try: self._oldsock.bind((server, 0))
        except socket.error:
            if not server:
                try: socket.inet_pton(socket.AF_INET6, self.server)
                except socket.error: pass
                else: server = self.server
            if not server:
                try: socket.inet_pton(socket.AF_INET, self.server)
                except socket.error: pass
                else: server = self.server
            if not server:
                ips = []
                try:
                    for item in socket.getaddrinfo(self.server, None):
                        if item[0] in [socket.AF_INET, socket.AF_INET6] and item[1] == socket.SOCK_STREAM:
                            ip = item[4][0]
                            if ip not in ips: ips.append(ip)
                except socket.error: pass
                else: server = random.choice(ips)
        return server
       
    def parse(self, txt):
        """ parse the incoming string into an IRC event. """
        rawstr = str(txt)
        obj = Event()
        obj._bot = self
        obj.arguments = []
        arguments = rawstr.split()
        obj.origin = arguments[0]
        if obj.origin.startswith(":"):
            obj.origin = obj.origin[1:]
            if len(arguments) > 1: obj.cmnd = arguments[1]
            if len(arguments) > 2:
                    txtlist = []
                    adding = False
                    for arg in arguments[2:]:
                        if arg.startswith(":"): adding = True ; txtlist.append(arg[1:]) ; continue
                        if adding: txtlist.append(arg)
                        else: obj.arguments.append(arg)
                    obj.txt = " ".join(txtlist)
        else: obj.cmnd = obj.origin ; obj.origin = self.server
        try: obj.nick, obj.userhost = obj.origin.split("!")
        except: pass
        if obj.arguments: obj.target = obj.arguments[-1]
        else: obj.target = ""
        if obj.target.startswith("#"): obj.channel = obj.target
        if not obj.txt and len(arguments) == 1: obj.txt = arguments[1]
        if not obj.txt: obj.txt = rawstr.split(":")[-1]
        return obj

    def announce(self, txt):
        """ announce txt on the registered channels. """
        for channel in self.channels: self._outqueue.put((channel, txt))

    def say(self, channel, txt):
        """ say text on a specific channel. """
        self._outqueue.put((channel, txt))
 
    def output(self, channel, line):
        """ split text into max. 450 chars lines and send that to the IRC server. """
        for txt in split_txt(line, 450):
            if time.time() - self._last < 3.5: time.sleep(3.5) 
            self.privmsg(channel, txt)

    def out(self, txt):
        """ the lowlevel function to output text to the server. """
        if not txt.endswith("\r\n"): txt += "\r\n"
        txt = txt[:512]
        txt = bytes(txt, "utf-8")
        logging.info("> %s" % txt)
        self._last = time.time()
        try: self._sock.send(txt)
        except AttributeError: self._sock.write(txt)

    def close(self):
        """ close the bot's sockets """
        logging.warn("# closing")
        if 'ssl' in self and self['ssl']: self.oldsock.shutdown(1) ; self.oldsock.close()
        else: self._sock.shutdown(1) ; self._sock.close()
        self.fsock.close()

    def join(self, channel, password=""):
        """ join a channel. """
        logging.warn("# join %s" % channel) 
        if password: self.out('JOIN %s %s' % (channel, password))
        else: self.out('JOIN %s' % channel)

    def joinall(self):
        """ join all registered channels """
        for channel in self.channels: self.join(channel)

    def part(self, channel):
        """ leave a channel. """
        self.out('PART %s' % channel)
        if channel in self.channels:
            self.channels.remove(channel)
            self.save()

    def noticed(self, event):
        """ called when the bot is noticed. """
        pass

    def connected(self, event):
        """ send usermodes and join all channels. """
        if "servermodes" in self: self.out("MODE %s %s" % (self.nick, self.servermodes))
        logging.warn("# connected %s:%s" % (self.server, self.port))
        self.joinall()

    def invited(self, event):
        """ join a channel on invitation. """
        self.join(event.channel)

    def joined(self, event):
        """ register the joined channel with the bot. """
        if event.channel and event.channel not in self.channels: self.channels.append(event.channel)

    def errored(self, event):
        """ print a log message on error. """
        logging.error("^ error %s %s" %  (event.origin, event.txt))
        self._error = event.origin
        self._status = "error"

    def pinged(self, event):
        """ ping callback. """
        self.pongcheck = True
        self.pong(event.txt)

    def ponged(self, event):
        """ pong callback. """
        self.pongcheck = False

    def quited(self, event):
        """ quit the bot. """
        if "Ping timeout" in event.txt and event.nick == self.nick: self.connecting()

    def privmsged(self, event):
        """ start DCC, CTCP responce or exec a command. """
        if event.txt.startswith("\001DCC"): self.dccconnect(event) ; return
        elif event.txt.startswith("\001VERSION"): self.ctcpreply(event.nick, "VERSION MEDS #%s - http://pypi.python.org/pypi/meds" % __version__) ; return
        event.cc = self.cc
        kernel.put(event)

    def ctcped(self, event):
        """ CTCP callback. """
        pass

    def h366(self, event):
        """ 366 callback. """
        pass 

    def h433(self, event):
        """ 433 callback. """
        self.donick(event.target + "_")
      
    def h513(self, event):
        """ 513 callback. """
        self.out("PONG %s" % event.txt.split()[-1])

    def donick(self, name):
        """ set bot's nick. """
        self.out('NICK %s\n' % name[:16]) ; self.nick = name

    def who(self, channel):
        """ ask who is on a channel. """
        self.out('WHO %s' % channel)

    def names(self, channel):
        """ query all names on a channel. """
        self.out('NAMES %s' % channel)

    def whois(self, nick):
        """ ask WHOIS data of a user. """
        self.out('WHOIS %s' % nick)

    def privmsg(self, channel, txt):
        """ send privmsg to the server. """
        self.out('PRIVMSG %s :%s' % (channel, txt))

    def voice(self, channel, nick):
        """ give voice permissions to a user. """
        self.out('MODE %s +v %s' % (channel, nick))

    def doop(self, channel, nick):
        """ give op permissions to a user. """
        self.out('MODE %s +o %s' % (channel, nick))

    def delop(self, channel, nick):
        """ remove op permissions of a user. """
        self.out('MODE %s -o %s' % (channel, nick))

    def quit(self, reason='https://pikacode.com/bart/meds'):
        """ send a quit message. """
        self.out('QUIT :%s' % reason)

    def notice(self, channel, txt):
        """ notice a channel. """
        self.out('NOTICE %s :%s' % (channel, txt))

    def ctcp(self, nick, txt):
        """ ctcp txt to a user, """
        self.out("PRIVMSG %s :\001%s\001" % (nick, txt))

    def ctcpreply(self, channel, txt):
        """ give a ctcp reply. """
        self.out("NOTICE %s :\001%s\001" % (channel, txt))

    def action(self, channel, txt):
        """ /me ;] """
        self.out("PRIVMSG %s :\001ACTION %s\001" % (channel, txt))

    def getchannelmode(self, channel):
        """ query channel modes. """
        self.out('MODE %s' % channel)

    def settopic(self, channel, txt):
        """ set TOPIC. """
        self.out('TOPIC %s :%s' % (channel, txt))

    def ping(self, txt):
        """ send PING. """
        self.out('PING :%s' % txt)

    def pong(self, txt):
        """ send PONG. """
        self.out('PONG :%s' % txt)

    def dcced(self, event, s):
        """" start DCC session. """
        s.send(bytes('Welcome to MEDS ' + event.nick + " !!\n", self.encoding))
        self.launch(self.dccloop, event, s)

    def dccloop(self, event, s):
        """ DCC loop. """
        sockfile = s.makefile('rw')
        s.setblocking(True)
        self.add(s)
        while 1:
            try:
                res = sockfile.readline()
                if not res: break
                res = res.rstrip()
                logging.info("< %s %s" % (event.origin, res))
                e = Event()
                e._bot = self
                e.txt = res
                e.outer = sockfile
                e.origin = event.origin
                kernel.put(e)
            except socket.timeout: time.sleep(0.01)
            except socket.error as ex:
                if ex.errno in [socket.EAGAIN, ]: continue
                else: raise
            except Exception as ex: logging.error(get_exception())
        sockfile.close()

    def dccconnect(self, event):
        """ connect back to DCC user. """
        event.parse()
        try:
            addr = event.args[2] ; port = event.args[3][:-1]
            port = int(port)
            if re.search(':', addr): s = socket.socket(socket.AF_INET6, socket.SOCK_STREAM)
            else: s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            s.connect((addr, port))
        except Exception as ex: logging.error(get_exception()) ; return
        self.dcced(event, s)
