# meds/event.py
#
#

""" event passed on to commands. """

from meds.errors import ENOTSET, ENODATE, ENOFUNC
from meds.utils.trace import get_exception
from meds.utils.url import strip_html
from meds.utils.getters import days
from meds.utils.tijd import to_day
from meds.object import Object
from meds.utils.join import j

import logging
import types
import time
import os

class Event(Object):

    """ Basic Event class. """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._func = None
        self.type = self.__class__.__name__
        self.threaded = False
        self.parsed = False
        self.iscmnd = False
        self.prefix = ""
        self.cmnd = ""
        self.rest = ""
        self.args = []
        self.fields = []
        self.default = False
        self.want = Object()
        self.notwant = Object()
        self.switch = Object() 
        self.boundaries = Object()
        self.created = time.ctime(time.time())
        self.done = False

    def __getattr__(self, name):
        try: return super().__getattr__(name)
        except AttributeError: pass
        try: return self[name]
        except: return ""

    def announce(self, txt):
        """ announce text. """
        from meds.kernel import kernel
        kernel.announce(txt)

    def dispatch(self):
        """ dispatch on the self._func attribute, if available. """
        if self._func: self._func(self)
        
    def join(self, *args, **kwargs):
        """ join thread that has been started on handling this event. """
        return self._thr.join(*args, **kwargs)

    def reply(self, txt):
        """ reply to origin of event. """
        from meds.kernel import kernel
        if type(txt) in [list, tuple]: txt = ",".join([str(x) for x in txt])
        if "outer" in self:
            self.outer.write(str(txt) + "\n")
            self.outer.flush()
            return
        else: self.say(self.channel, txt)
        
    def say(self, channel, txt):
        """ use _bot to output txt. """
        if "_bot" not in self: raise ENOTSET("_bot")
        self._bot.say(channel, txt)
        self._last = time.time()

    def ok(self, txt=""):
        """ reply with 'ok'. """
        self.reply("ok %s" % txt)

    def display(self, obj, keys=[], txt=""):
        """ display an object, showing the provided keys, if the are available. """
        res = ""
        if not keys: keys = obj.keys()
        for key in sorted(keys):
            val = obj.get(key, None)
            if val: res += " " + str(val)
        try: res += " - %s" % days(obj)
        except ENODATE: pass
        txt += res
        if txt: self.reply(txt.strip()) ; return True
        return False

    def parse(self, txt=""):
        """ parse a string into an Event. """
        from meds.kernel import kernel
        if self.parsed: return
        if txt: self.txt = txt.rstrip()
        if self.txt.endswith("&"): self.threaded = True ; self.txt = self.txt[:-1]
        splitted = self.txt.split()
        c = 0
        for word in splitted:  
            if not c:
                if word[0] == "!": word = word[1:] ; self.cmnd = word.lower()
                else: self.cmnd = word.lower()
                c += 1
                continue   
            if word.startswith("+"):
                try: self.karma = int(word)
                except ValueError: self.karma = 0
            if "http" in word: self.args.append(word) ; self.rest += " " + word ; continue
            try:
                key, value = word.split("=", 1)
                if key == "start": self.boundaries.start = to_day(value) ; continue
                if key == "end": self.boundaries.end = to_day(value) ; continue
                pre = key[0]
                op = key[-1] 
                post = value[-1]
                last = word[-1] 
                if key == "i":  
                    try: self.index = int(value)
                    except: pass
                if post == "-": value = value[:-1]
                if key.startswith("!"): key = key[1:] ; self.switch[key] = value ; continue
                if op == "-": key = key[:-1] ; self.notwant[key] = value
                else: self.want[key] = value
                if post == "-" : continue
                self.fields.append(key)
                self.args.append(key)
                self.rest += " " + key
            except (IndexError, ValueError) as ex:
                self.fields.append(word)
                self.args.append(word)
                self.rest += " " + word
        self.rest = self.rest.strip()  
        self.parsed = True
        return self
