# meds/kernel.py
#
#

""" runtime namespace. """

from meds.errors import ESET, ENOTSET, ERESERVED
from meds.utils.misc import exclude, include, locked
from meds.utils.trace import get_exception
from meds.utils.name import name
from meds.scheduler import Scheduler
from meds.launcher import Launcher
from meds.storage import Storage
from meds.event import Event
from meds.object import Object
from meds.utils.join import j
from meds.bots.cli import CLI
from meds.bots import Bot
from meds.ool import OOL
from meds.cfg import cfg

import importlib
import logging
import pkgutil
import select
import types
import sys
import os

class Kernel(Scheduler, Launcher, Storage):

    """ bot's kernel. """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.booted = False
        self.bot = CLI()		# command line interface bot
        self.cfg = cfg			# config copy
        self.cts = Object()		# counters
        self.cbs = OOL()		# callbacks
        self.cmnds = OOL()		# commands
        self.fleet = OOL()		# bots
        self.names = OOL()		# module names
        self.objs = Object()		# runtime objects
        self.table = Object()		# command table

    def dispatch(self, event):
        """ launch a thread to handle an event. """
        self._nr += 1
        if not event.parsed: event.parse()
        if event.txt and event.cc and event.cc != event.txt[0]: return
        if event.cmnd in ["init", "shutdown"]: logging.warn("ERESERVED %s" % event.cmnd) ; return []
        logging.info("# dispatch %s %s" % (event.origin, event.cmnd))
        if not self.names: self.boot()
        names = self.names.get(event.cmnd, [])
        for name in names:
            if not self.table.get(name, None): self.table[name] = self.direct(name)
        cmnds = self.cmnds.get(event.cmnd, [])
        thrs = []
        for cmnd in cmnds:
            event.iscmnd = True
            thr = self.launch(cmnd, event)
            thrs.append(thr)
        return thrs

    def callcb(self, event):
        for cb, funcs in self.cbs.items():
            for func in funcs:
                try: func(event)
                except: get_exception()

    def boot(self):
        """ initialize the kernel. """
        from meds.event import Event
        self.walk()
        if cfg.main.resume:
            try: kernel = Object().load(j(cfg.main.workdir, "kernel")) ; self.resume(kernel)
            except: logging.warn("# no kernel dump found") 
            logging.warn("# resume %s" % kernel._path)
        if cfg.main.init:
            thrs = self.initialize()
            if thrs: self.waiter(thrs)
            self.save(j(cfg.main.workdir, "kernel"))
        self.booted = True
        self.fleet.register("CLI", self.bot)
        self.ready()

    def start(self):
        if cfg.main.shell: self.boot()
        self.launch(super().start)
        self.bot.start()
        
    @locked
    def shutdown(self, close=True):
        for bots in self.fleet.values():
            for bot in bots:
                try: bot.stop()
                except: logging.error(get_exception())

    def resume(self, kernel): pass

    def initialize(self):
        self.walk(["meds.run"])
        thrs = []
        event = Event()
        event.parse()
        for name in self.modules(["meds.run"]):
            n = name.split(".")[-1]
            mod = self.table.get(name, None)
            if not mod: self.table[name] = mod = self.direct(name)
            if cfg.main.all or n in cfg.main.init and n not in cfg.main.exclude:
                thr = self.launch(mod.init, event)
                thrs.append(thr)
        return thrs

    def direct(self, name, package=None):
        """ directly import a module. """
        logging.info("! load %s" % name)
        return importlib.import_module(name, package)

    def reload(self, name, force=False):
        """ reload/load a module. """
        from meds.event import Event
        e = Event()
        n = name.split(".")[-1]
        self.unload(name)
        if name in self.table and "shutdown" in dir(self.table[name]): self.table[name].shutdown(e) ; e.ready()
        if name not in self.table or force: self.table[name] = self.direct(name)
        else: self.table[name] = importlib.reload(self.table[name])
        if "init" in dir(self.table[name]): self.table[name].init(e) ; e.ready()
        return self.table[name]

    def unload(self, name):
        if name in self.cmnds:
            for key in self.cmnds[name]: del self.cmnds[key]
        if name in self.names: del self.names[name]

    def walk(self, packages=[], force=False):
        """ load all modules. """
        thrs = []
        for modname in sorted(list(self.modules(packages)), reverse=True):
            mod = self.direct(modname)
            if force: self.table[modname] = mod
            for key in dir(mod):
                obj = getattr(mod, key, None)
                if obj and type(obj) == types.FunctionType:
                    if key.startswith("cb_"):
                        logging.warn("# add calback %s" % name(obj))
                        self.cbs.register(key, obj)
                    elif "event" in obj.__code__.co_varnames:
                        self.names.register(key, modname)
                        self.cmnds.register(key, obj)

    def list(self, want=""):
        """ list all functions in modules loaded with the bot. """
        for name in self.modules():
            if not include(name): continue
            if want and want not in name: continue
            mod = self.direct(name)
            for key in dir(mod):
                if key in ["init", "shutdown"]: continue
                obj = getattr(mod, key, None)
                if obj and type(obj) == types.FunctionType:
                    if "event" in obj.__code__.co_varnames: yield key

    def modules(self, packages=[]):
        """ look for all modules that can be loaded. """
        for name in packages or cfg.main.packages:
            package = self.direct(name)
            for pkg in pkgutil.walk_packages(package.__path__, name + "."): yield pkg[1]

    def bots(self):
        for bots in self.fleet.values():
            for bot in bots: yield bot

    def out(self, txt): self.announce(txt)

    def announce(self, txt):
        for bots in self.fleet.values():
            for bot in bots: bot.announce(txt)

    def find(self, cmnd):
        """ command lookup. """
        names = self.names.get(cmnd, None)
        if names:
            for name in names:
                mod = self.table[name]
                obj = getattr(mod, cmnd, None)
                if obj and type(obj) == types.FunctionType:
                    if "event" in obj.__code__.co_varnames: yield obj

    def update(self, obj):
        for key in obj:
            val = self.get(key, None)
            if not val: continue
            typed = type(val)
            if not typed: continue
            if " at " in obj[key]: continue
            self[key] = typed(obj[key])

kernel = Kernel()
