# object.py
#
#

""" dict with dotted name access. """

from meds.utils.signature import make_signature, verify_signature
from meds.utils.misc import locked, headertxt
from meds.utils.getters import path as dopath
from meds.utils.getters import  urled, slice
from meds.utils.json import dumps
from meds.utils.tijd import rtime
from meds.utils.file import cdir
from meds.utils.join import j
from meds.errors import ESET, ENOJSON

from collections import UserDict, defaultdict
import threading
import logging
import inspect
import random
import fcntl
import json
import time
import os

class Object(dict):

    """ 
        An Object adds 'dotted access' to dicts to allow for better readable code. Instead of brackets one can now use . to access and set attributes.
        Object's also allow for saving JSON to disk to allow it to be restored at a later time.

    """

    def __getattribute__(self, name):
        try: val = self[name]
        except KeyError: val = super().__getattribute__(name)
        return val

    def __getattr__(self, name):
        if name == "url": self["url"] = urled(self)
        if name == "modname": return self.__class__.__module__
        if name == "type": return self.__class__.__name__
        if name == "_ready":  self._ready = threading.Event()
        if name == "_nr": self._nr = 0 
        if name not in self: raise AttributeError(name)
        return self[name]

    def __str__(self): return self.nice()

    def __repr__(self):
        return '<%s.%s at %s>' % (
            self.__class__.__module__,
            self.__class__.__name__,
            hex(id(self))
        )

    def __setitem__(self, name, value):
        return dict.__setitem__(self, name, value)

    def __setattr__(self, name, value):
        try:
            val = super().__getattribute__(name)
            if inspect.ismethod(val): raise AttributeSet
        except: pass
        return dict.__setitem__(self, name, value)

    def __contains__(self, name):
        try: self[str(name)] ; return True
        except KeyError: return super().__contains__(str(name))

    def consume(self, value):
        try:
            nr = 1
            for x in value: self["key%s" % nr] = value
            return True
        except: pass
        try: self.update(value) ; return True
        except: pass
        return False

    def register(self, key, val):
        if key in dir(self): raise ESET(key)
        logging.info("# register %s %s" % (key, str(val)))
        self[key] = val   

    def random(self):
        key = random.choice(list(self.keys()))
        return self[key]

    def grep(self, name):
        res = []
        for key, value in self.items():
            if name in str(value) and value not in res: res.append(value)
        return res

    def merge(self, obj):
        """ type: Object -> None - merge another Object into this one, not overriding already existing entries. """
        for key in obj:
            if key not in self: self[key] = obj[key]

    def search(self, name):
        """ search the Object for keys, skipping keys that start with an underscore. """
        for key in self.keys():
            if key.startswith("_"): continue
            if key == name: yield key
            elif name == key.split(".")[-1]: yield key

    def nice(self, *args, **kwargs):
        """ return a prettied JSON string of the Object. """
        return dumps(self, indent=4, sort_keys=True)

    def json(self, *args, **kwargs):
        """ return JSON string. """
        return dumps(self, *args, **kwargs)

    def load(self, path=""):
        """ construct an Object from the provided path, or Object._path if the Object has already been saved. """
        if not path: path = self._path
        ondisk = self.read(path)
        fromdisk = json.loads(ondisk)
        try: verify_signature(fromdisk["data"], fromdisk["signature"]) 
        except: logging.info("ESIGNATURE on %s" % path)
        try: self.update(fromdisk["data"])
        except: logging.info("# EDATA %s" % path) ; self.update(fromdisk)
        if "saved" in fromdisk: self._saved = fromdisk["saved"]
        self._path = path
        return self

    def read(self, path):
        """ read data from disk, skipping comments. """
        f = open(path, "r", encoding="utf-8")
        res = ""
        for line in f:
            if not line.strip().startswith("#"): res += line
        if not res.strip(): return "{}"
        f.close()
        return res

    def prepare(self):
        """ prepare the Object for saving. """
        todisk = Object()
        todisk.data = slice(self)
        todisk.type = self.__class__.__name__
        todisk.data.saved = time.ctime(time.time())
        todisk.data.url = urled(self)
        try: todisk.signature = make_signature(todisk["data"])
        except: pass
        try: result = dumps(todisk, indent=4, ensure_ascii=False, sort_keys=True)
        except TypeError: raise ENOJSON()
        return result

    def save(self, stime=""):
        """ save the Object to disk, optional day/time stamp can be provided as part of the filename. """
        if not stime: stime = rtime()
        path = j(dopath(self), stime)
        self.sync(path)
        return stime

    @locked
    def sync(self, path=""):
        """synchronise an already existing Object."""
        if not path:
            try: path = self._path
            except AttributeError: pass
        if not path: path = j(dopath(self), rtime())
        self._path = path
        logging.warn("# sync %s" % path)
        d, fn = os.path.split(path)
        cdir(d)
        todisk = self.prepare()
        datafile = open(os.path.abspath(path) + ".tmp", 'w')
        fcntl.flock(datafile, fcntl.LOCK_EX | fcntl.LOCK_NB)
        datafile.write(headertxt % "%s characters" % len(todisk))
        datafile.write(todisk)
        datafile.write("\n")
        fcntl.flock(datafile, fcntl.LOCK_UN)
        datafile.flush()
        datafile.close()
        os.rename(path + ".tmp", path)
        return path

    def isSet(self): return self._ready.isSet()

    def ready(self): logging.debug("! ready %s" % repr(self)) ; self._ready.set()

    def clear(self): self._ready.clear() ; logging.debug("! clear %s" % repr(self))

    def wait(self, sec=180.0): logging.debug("! wait %s" % repr(self)) ; self._ready.wait(sec)
